using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Runtime.InteropServices;


namespace ArcRen
{
    static class ArchiveFactory
    {
        static public Archive Factory(string arcFileName)
        {
            string ext = Path.GetExtension(arcFileName).ToLower();
            switch (ext)
            {
                case ".zip": return new ZipArchive(arcFileName);
                case ".lzh": return new LhaArchive(arcFileName);
                case ".cab": return new CabArchive(arcFileName);
                default: break;
            }
            return null;
        }
    }

    /// <summary>
    /// ɂNX
    /// </summary>
    abstract class Archive : IDisposable
    {
        #region Win32API
        [DllImport("kernel32")]
        protected extern static int LoadLibrary(string lpLibFileName);
        [DllImport("kernel32")]
        protected extern static bool FreeLibrary(int hLibModule);
        [DllImport("kernel32")]
        protected extern static int GetProcAddress(int hModule, string lpProcName);
        #endregion

        #region ArchiverAPI (Invoke)
        [DllImport("Invoke", EntryPoint="InvokeFunc")]
        protected extern static UInt16 InvokeGetVersion(int funcptr);
        [DllImport("Invoke", EntryPoint="InvokeFunc")]
        protected extern static bool InvokeGetRunning(int funcptr);
        [DllImport("Invoke", EntryPoint="InvokeFunc")]
        protected extern static bool InvokeCheckArchive(int funcptr, string szFileName, int iMode);
        [DllImport("Invoke", EntryPoint = "InvokeFunc")]
        protected extern static int InvokeMain(int funcptr, int hwnd, string szCmdLine, StringBuilder szOutput, int dwSize);
        [DllImport("Invoke", EntryPoint = "InvokeFunc")]
        protected extern static int InvokeOpenArchive(int funcptr, int hwnd, string szFileName, int dwMode);
        [DllImport("Invoke", EntryPoint = "InvokeFunc")]
        protected extern static int InvokeCloseArchive(int funcptr, int harc);
        [DllImport("Invoke", EntryPoint = "InvokeFunc")]
        protected extern static int InvokeFindFirst(int funcptr, int harc, string szWildName, IntPtr lpSubInfo);
        [DllImport("Invoke", EntryPoint = "InvokeFunc")]
        protected extern static int InvokeFindNext(int funcptr, int harc, IntPtr lpSubInfo);
        [DllImport("Invoke", EntryPoint = "InvokeFunc")]
        protected extern static int InvokeGetFileName(int funcptr, int harc, StringBuilder lpBuffer, int nSize);
        [DllImport("Invoke", EntryPoint = "InvokeFunc")]
        protected extern static int InvokeGetAttribute(int funcptr, int harc);

        protected int ptrGetVersion = 0;
        protected int ptrGetRunning = 0;
        protected int ptrCheckArchive = 0;
        protected int ptrMain = 0;
        protected int ptrOpenArchive = 0;
        protected int ptrCloseArchive = 0;
        protected int ptrFindFirst = 0;
        protected int ptrFindNext = 0;
        protected int ptrGetFileName = 0;
        protected int ptrGetAttribute = 0;
        #endregion

        protected int hModule = 0;  // DLL̃nh

        protected string arcFileName;  // Ƀt@C


        public Archive(string arcName, string libName, string funcName)
        {
            hModule = LoadLibrary(libName);
            if (hModule == 0)
            {
                throw new ApplicationException(libName + "[hł܂ł");
            }

            // A[JCoDLL̊֐̃AhX擾
            ptrGetVersion = GetProcAddress(hModule, funcName + "GetVersion");
            ptrGetRunning = GetProcAddress(hModule, funcName + "GetRunning");
            ptrCheckArchive = GetProcAddress(hModule, funcName + "CheckArchive");
            ptrMain = GetProcAddress(hModule, funcName);
            ptrOpenArchive = GetProcAddress(hModule, funcName + "OpenArchive");
            ptrCloseArchive = GetProcAddress(hModule, funcName + "CloseArchive");
            ptrFindFirst = GetProcAddress(hModule, funcName + "FindFirst");
            ptrFindNext = GetProcAddress(hModule, funcName + "FindNext");
            ptrGetFileName = GetProcAddress(hModule, funcName + "GetFileName");
            ptrGetAttribute = GetProcAddress(hModule, funcName + "GetAttribute");

            if (!File.Exists(arcName))
            {
                throw new ApplicationException(arcName + "݂͑܂");
            }
            arcFileName = arcName;

        }

        public void Dispose()
        {
            if (hModule != 0)
            {
                FreeLibrary(hModule);
                hModule = 0;
            }
        }

        public UInt16 GetVersion()
        {
            return InvokeGetVersion(ptrGetVersion);
        }

        public List<string> GetFileList()
        {
            List<string> fileList = new List<string>();

            int harc = InvokeOpenArchive(ptrOpenArchive, 0, arcFileName, 0);
            if (harc == 0)
            {
                throw new ApplicationException("ɂJ܂ł");
            }

            if (InvokeFindFirst(ptrFindFirst, harc, "*.*", IntPtr.Zero) != -1)
            {
                do
                {
                    StringBuilder strFileName = new StringBuilder(1024);
                    if (InvokeGetFileName(ptrGetFileName, harc, strFileName, 1024) == 0)
                    {
                        fileList.Add(strFileName.ToString());
                    }
                }
                while (InvokeFindNext(ptrFindNext, harc, IntPtr.Zero) != -1);
            }

            InvokeCloseArchive(ptrCloseArchive, harc);

            return fileList;
        }

    }

    class ZipArchive : Archive
    {
        public ZipArchive(string arcFileName)
            : base(arcFileName, "unzip32.dll", "UnZip")
        {
        }
    }

    class LhaArchive : Archive
    {
        public LhaArchive(string arcFileName)
            : base(arcFileName, "unlha32.dll", "Unlha")
        {
        }
    }

    class CabArchive : Archive
    {
        public CabArchive(string arcFileName)
            : base(arcFileName, "cab32.dll", "Cab")
        {
        }
    }
}
